//=============================================================================================================
//                                           I N T R O D U C T I O N
//=============================================================================================================
//
//  This software is a patch of the existing bootloader feature inside LPC1800 / LPC4300.
//
//  It solves the following application problem:
//    How can I enter USB0 DFU bootmode from application code without physically modifying the bootmode pins ? 
//   
//  The bootmode pins P1_1, P1_2, P2_8 and P2_9 determine the way the bootloader behaves. This is described in 
//  detail in the user manual. If it's not possible to change the bootmode pins, the MCU will always start from 
//  the selected bootsource. For e.g. a firmware upgrade it is very useful to set a different boot source from 
//  the application code instead of performing a manual bootmode pin modification.
//
//  The patch uses the Flash Patch and Breakpoint unit of the Cortex-M to modify the existing boot ROM code.
//
//  As this is a patch, NXP gives no warranty on the final functionality. The implementation and test is up to
//  the end product manufacturer.
//  NXP will not provide source code or any additional information about the bootloader in LPC1800/4300.
//
//  This is a small stand-alone example which does not fit into any software package concept from NXP or 3rd party.
//
//  Supported MCUs: 
//
//   - LPC1850/30/20/10 Rev A                                     Bootcode v11.2
//   - LPC4350/30/20/10 Rev A                                     Bootcode v11.2
//   - LPC4370                                                    Bootcode v11.2
//   - LPC1857 and all other flash based derivatives rev "-"      Bootcode v12.1
//   - LPC4357 and all other flash based derivatives rev "-"      Bootcode v12.1
//   - LPC1857 and all other flash based derivatives rev "A"      Bootcode v12.2  (Q4/2014)
//   - LPC4357 and all other flash based derivatives rev "A"      Bootcode v12.2  (Q4/2014)

//
//=============================================================================================================

#ifdef CORE_M4
  #include "LPC43xx.h"
#endif

#ifdef CORE_M3
  #include "LPC18xx.h"
#endif


// Define registers of the Flash Patch and Breakpoint unit (FPB).
// This is part of the Cortex-M3/M4 core and is described in detail in the ARMv7-M Architecture Reference Manual
typedef struct {
    __IO uint32_t CTRL;
    __IO uint32_t REMAP;
    __IO uint32_t CODE_COMP[6];
    __IO uint32_t LIT_COMP[2];
} FPB_Type;

#define FPB ((FPB_Type *)0xE0002000)

// The FPB contains both a global enable and individual enables for the eight comparators. If the
// comparison for an entry matches, the address is remapped to the address set in the remap register 
// plus an offset corresponding to the comparator that matched.



//#############################################################################################################
//                                                M A I N
//#############################################################################################################
int main (void)
{
  volatile int i;
 
	// This delay is just for debug purposes, in order to get access with a debugger before 
	// more "dreadful things" could happen ;-)
  for (i = 0; i < 10000000; i++);
	
	
  //=============================================================================================================
	//                                           T U T O R I A L
	//=============================================================================================================
	//
	// STEP #1:  Wind back MCU to a state which is similar to the system state after a normal bootloader run.
	//           This could be pretty application specific, therefore the implementation and test is individual.
	//           But the following 3 steps need to be integrated into your application code.
	// 
	// Example:  - set the PLL back to 96 MHz (bootloader also sets it to 96MHz)
	//           - reset as many peripherals as possible.
  //             LPC_RGU->RESET_CTRL0 = 0x387F1200;   // this works for LPC1857/37/33 etc and LPC4357/37/33 etc in rev. "-"
  //             LPC_RGU->RESET_CTRL1 = 0x17FFF7FF;   // this works for LPC1857/37/33 etc and LPC4357/37/33 etc in rev. "-"
  
	
  // Disable or wind back settings of your application code
	// < code here>
   
	// Reset as many peripherals as possible in LPC4357 and LPC1857
	LPC_RGU->RESET_CTRL0 = 0x387F1200;  // this should work for LPC1857/37/33 etc and LPC4357/37/33 etc with bootcode v12.1
  LPC_RGU->RESET_CTRL1 = 0x17FFF7FF;  // this should work for LPC1857/37/33 etc and LPC4357/37/33 etc with bootcode v12.1

  // Resets for LPC4350 and LPC1850 (bootcode v11.2) needs to be tested	
	// Start with a single bit and then add bit by bit.
	// Doing no reset at least seems to work in a first step.
  //LPC_RGU->RESET_CTRL0 = 0x????????; 
  //LPC_RGU->RESET_CTRL1 = 0x????????; 


  //=============================================================================================================
	//                                           T U T O R I A L
	//=============================================================================================================
	//
	// STEP #2:  Set up the Flash Patch and Breakpoint unit (FPB) in such a way that the result from a bootmode pin
	//           query is faked. The level on the boot pins are latched after reset by hardware, so there's nothing 
	//           you can do here. But the bootloader will ask for the result in one of its functions, and this is 
	//           where we jump in: if the bootloader comes to a specific address (the one we store in the 
	//           FPB CODE_COMP register) the ARM will fetch the code from the SRAM location we defined in the REMAP 
	//           register. The code in this case is pretty simple, we just store 0x15 in R0, the effect is simply 
	//           that we replaced the true level of the bootcode pins by a faked level, 0x15 means USB0 DFU bootmode.
	//
	// Note:  Please keep in mind that a debugger also uses the FPB to set breakpoints (also "Run to Cursor" creates
	//        a breakpoint). So if you think you need to debug into this code, then please be aware that the 
	//        debugger will change the FPB settings without letting you know ;-)
	
	
  // Set SRAM location for patch code
	FPB->REMAP = 0x20000000;         // REMAP table @ 0x20000000 
	
	// Define the patch code instructions
  ((volatile uint32_t *)0x20000000)[0] = 0x20152015;  // movs r0,#0x15 (2x)
	
	// Set the address comperator  (must be an odd address)
	// Required address from boot ROM v12.1  (e.g. LPC1857 rev. "-")
  FPB->CODE_COMP[0] = 0x10401A04 | 0x00000001; 

  // Required address from boot ROM v11.2  (e.g. LPC1850 rev. A)
  //FPB->CODE_COMP[0] = 0x10400F64 | 0x00000001; 

  // Required address from boot ROM v12.2   (e.g. LPC1857 rev. A)
  //FPB->CODE_COMP[0] = 0x10402404 | 0x00000001; 
	
	// Enable the FPB unit
  FPB->CTRL = 0x00000003;          


  //=============================================================================================================
	//                                           T U T O R I A L
	//=============================================================================================================
	//
	// STEP #3:   Jump to bootcode functions in ROM in order to end up in DFU bootmode.
	//            These are individual for each boot ROM version.
	//            In the last function tmBootRomISP_run the bootloader will enter, as required, the USB0 DFU bootmode
	//            and the board should appear in the device manager of a PC as "LpcDevice" 
	
	
  // Enable the clock for Random Number Generator inside the internal flash block. This is required for a correct 
	// operation of the bootloader code for the flash types LPC1857/4357.
	// This is NOT required for the LPC1850/4350, because there is no internal flash.
	// This bit is not documented in the LPC1857/4357 user manual.
  LPC_CREG->CREG6 &= ~(1u << 17);    // please comment out for LPC1850/4350         


  // Please remember what has been said before:
	// ==========================================
  // If you debug into this section using breakpoints or "Run to Cursor", then your settings of the FPB were  
	// changed by the debugger in the background, therefore the expected code replacement will not take place.
	
	// Addresses from boot ROM v12.1  (e.g. LPC1857 rev. "-")
  ((void(*)(int))0x1040214D)(LPC_OTP_BASE);   // Clocks_Init
  ((void(*)(int))0x104021C1)(LPC_OTP_BASE);   // tmBootRomConfig_Clocks_and_Pins
  ((void(*)(int))0x104038DF)(LPC_OTP_BASE);   // tmBootRomISP_init
  ((void(*)(int))0x104038F3)(LPC_OTP_BASE);   // tmBootRomISP_run
/*	
	// Addresses from ROM 11.2  (e.g. LPC1850 rev. A)
  ((void(*)(int))0x104014B9)(LPC_OTP_BASE);   // Clocks_Init
  ((void(*)(int))0x10400A05)(LPC_OTP_BASE);   // tmBootRomSrc_init
  ((void(*)(int))0x10400A27)(LPC_OTP_BASE);   // tmBootRomSrc_start
	
	// Addresses from boot ROM v12.2  (e.g. LPC1857 rev. A)
  ((void(*)(int))0x10402B4D)(LPC_OTP_BASE);   // Clocks_Init
  ((void(*)(int))0x10402D43)(LPC_OTP_BASE);   // tmBootRomConfig_Clocks_and_Pins
  ((void(*)(int))0x10405563)(LPC_OTP_BASE);   // tmBootRomISP_init
  ((void(*)(int))0x10405577)(LPC_OTP_BASE);   // tmBootRomISP_run
*/
  
	while (1)  // we will never come here if everything worked as planned
  ;
}


